Namespace BinningData

    Friend Class BinningProgram

        Shared Sub Main(ByVal args() As String)

            Try

                Console.WriteLine(vbCrLf & "Begin discretization of continuous data demo" & vbCrLf)

                Dim rawData() As Double = {66, 66, 66, 67, 67, 67, 67, 68, 68, 69, 73, 73, 73, 74, 76, 78, 60, 61, 62, 62}

                Console.WriteLine("Raw data:")
                ShowVector(rawData, 2, 10)

                Console.WriteLine(vbCrLf & "Creating a discretizer on the raw data")
                Dim d As New Discretizer(rawData)
                Console.WriteLine(vbCrLf & "Discretizer creation complete")

                Console.WriteLine(vbCrLf & "Displaying internal structure of the discretizer:" & vbCrLf)
                Console.WriteLine("-------------------------------------------------------------" & vbCrLf)
                Console.WriteLine(d.ToString)
                Console.WriteLine(vbCrLf & "-------------------------------------------------------------")

                Console.WriteLine(vbCrLf & "Generating three existing and three new data values")
                Dim newData() As Double = {62.0, 66.0, 73.0, 59.5, 75.5, 80.5}

                Console.WriteLine(vbCrLf & "Data values:")
                ShowVector(newData, 2, 10)

                Console.WriteLine(vbCrLf & "Discretizing the data:" & vbCrLf)
                For i  = 0 To newData.Length - 1

                    Dim cat As Integer = d.Discretize(newData(i))
                    Console.WriteLine(newData(i).ToString("F2") & " -> " & cat)

                Next i

                Console.WriteLine(vbCrLf & vbCrLf & "End discretization demo")
                Console.ReadLine()

            Catch ex As Exception

                Console.WriteLine(ex.Message)
                Console.ReadLine()

            End Try

        End Sub ' Main

        Public Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal itemsPerRow As Integer)

            For i  = 0 To vector.Length - 1

                If i Mod itemsPerRow = 0 Then
                    Console.WriteLine("")
                End If
                Console.Write(vector(i).ToString("F" & decimals) & " ")

            Next i
            Console.WriteLine("")

        End Sub

    End Class ' Program

    Public Class Discretizer

        Private data() As Double ' sorted copy of distinct raw numeric data
        Private k As Integer ' number of clusters = number of categories. must be > data.Length
        Private means() As Double ' average of pts in each cluster, from sums] and counts[
        Private clustering() As Integer ' index = index into data[], value = cluster ID

        Public Sub New(ByVal rawData() As Double)

            Dim sortedRawData(rawData.Length - 1) As Double
            Array.Copy(rawData, sortedRawData, rawData.Length)
            Array.Sort(sortedRawData)
            Me.data = GetDistinctValues(sortedRawData)
            Me.clustering = New Integer(data.Length - 1) {}

            Me.k = CInt(Fix(Math.Sqrt(data.Length))) ' heuristic

            Me.means = New Double(k - 1) {}
            Me.Cluster()

        End Sub

        Private Shared Function GetDistinctValues(ByVal array() As Double) As Double()

            Dim distinctList As New List(Of Double)
            distinctList.Add(array(0))
            For i  = 0 To array.Length - 2
                If AreEqual(array(i), array(i + 1)) = False Then
                    distinctList.Add(array(i + 1))
                End If
            Next i

            Dim result(distinctList.Count - 1) As Double
            distinctList.CopyTo(result)
            Return result

        End Function

        Private Shared Function AreEqual(ByVal x1 As Double, ByVal x2 As Double) As Boolean

            If Math.Abs(x1 - x2) < 0.000001 Then
                Return True
            Else
                Return False
            End If

        End Function

        Public Function Discretize(ByVal x As Double) As Integer

            ' for any value x, compute distance to each cluster mean,
            ' return closest index, which is a cluster, which is the category
            Dim distances(k - 1) As Double
            For c  = 0 To k - 1
                distances(c) = Distance(x, means(c))
            Next c
            Return MinIndex(distances)

        End Function

        Public Overrides Function ToString() As String

            Dim s As String = ""
            s &= "Distinct data:"
            For i  = 0 To data.Length - 1

                If i Mod 10 = 0 Then
                    s &= vbCrLf
                End If
                s &= data(i).ToString("F2") & " "

            Next i
            s &= vbCrLf & vbCrLf & "k = " & k
            s &= vbCrLf & vbCrLf & "Clustering:" & vbCrLf
            For i  = 0 To clustering.Length - 1
                s &= clustering(i) & " "
            Next i
            s &= vbCrLf & vbCrLf & "Means:" & vbCrLf
            For i  = 0 To means.Length - 1
                s &= means(i).ToString("F2") & " "
            Next i
            Return s

        End Function

        ' ================================

        Private Sub InitializeClustering()

            Dim initialIndexes() As Integer = GetInitialIndexes()
            For di  = 0 To data.Length - 1

                Dim c As Integer = InitialCluster(di, initialIndexes)
                clustering(di) = c

            Next di

        End Sub

        Private Function GetInitialIndexes() As Integer()

            ' assumes data is sorted
            Dim interval As Integer = data.Length \ k
            Dim result(k - 1) As Integer
            For i  = 0 To k - 1
                result(i) = interval * (i + 1)
            Next i
            Return result

        End Function

        Private Function InitialCluster(ByVal di As Integer, ByVal initialIndexes() As Integer) As Integer

            ' for data index di (like 5) and a set of initial indexes (like 2,4,6) what cluster?
            For i  = 0 To initialIndexes.Length - 1
                If di < initialIndexes(i) Then
                    Return i ' assumes initialIndexes are ordered
                End If
            Next i
            Return initialIndexes.Length - 1 ' last cluster

        End Function

        Private Sub Cluster()

            InitializeClustering()
            ComputeMeans()

            Dim changed As Boolean = True
            Dim success As Boolean = True
            Dim ct As Integer = 0
            Dim maxCt As Integer = data.Length * 10 ' heuristic
            Do While changed = True AndAlso success = True AndAlso ct < maxCt

                ct += 1
                changed = AssignAll()
                success = ComputeMeans()

            Loop

        End Sub

        Private Function ComputeMeans() As Boolean

            ' return false if at some point a count goes to zero
            Dim sums(k - 1) As Double
            Dim counts(k - 1) As Integer

            For i  = 0 To data.Length - 1

                Dim c As Integer = clustering(i) ' cluster ID
                sums(c) += data(i)
                counts(c) += 1

            Next i

            For c  = 0 To sums.Length - 1

                If counts(c) = 0 Then
                    Return False ' serious algorithm problem
                Else
                    sums(c) = sums(c) / counts(c)
                End If

            Next c

            sums.CopyTo(Me.means, 0)
            Return True ' all means computed successfully

        End Function

        Private Function AssignAll() As Boolean

            ' scan data, assign all data (indexes) to a cluster, return true if any changes
            Dim changed As Boolean = False
            Dim distances(k - 1) As Double ' distance to each cluster mean
            For i  = 0 To data.Length - 1

                For c  = 0 To k - 1
                    distances(c) = Distance(data(i), means(c))
                Next c
                Dim newCluster As Integer = MinIndex(distances)
                If newCluster <> clustering(i) Then

                    changed = True
                    clustering(i) = newCluster

                End If

            Next i
            Return changed

        End Function

        Private Function MinIndex(ByVal distances() As Double) As Integer

            Dim indexOfMin As Integer = 0
            Dim smallDist As Double = distances(0)
            For k  = 0 To distances.Length - 1

                If distances(k) < smallDist Then

                    smallDist = distances(k)
                    indexOfMin = k

                End If

            Next k
            Return indexOfMin

        End Function

        Private Shared Function Distance(ByVal x1 As Double, ByVal x2 As Double) As Double

            Return Math.Sqrt((x1 - x2) * (x1 - x2))

        End Function

    End Class ' class

End Namespace ' ns
